@@ -1,6 +1,6 @@
-/*  
+/*
     Process viewer/killer utility
-    Copyright (c) Tudor Hulubei & Andrei Pitis, April 1994
+    Copyright (c) Tudor Hulubei & Andrei Pitis, May 1994
 
 This file is part of UIT (UNIX Interactive Tools)
 
@@ -41,10 +41,12 @@ UIT; see the file COPYING.  If not, write to the Free Software Foundation,
 
 int SCREEN_X, SCREEN_Y;
 
+#define MAX_KEYS        2048
 #define MAX_PROCESS	1024
 #define KILLER_FIELDS	  12
 
-static char PSFields[KILLER_FIELDS][40] = 
+
+static char PSFields[KILLER_FIELDS][40] =
 {
     "TitleForeground",
     "TitleBackground",
@@ -86,6 +88,7 @@ static int PSColors[KILLER_FIELDS] =
 #define StatusBrightness		PSColors[11]
 
 
+
 #ifdef HAVE_LINUX
 int LinuxConsole = 1;
 int ColorMonitor = 1;
@@ -94,6 +97,10 @@ int LinuxConsole = 0;
 int ColorMonitor = 0;
 #endif
 
+
+extern char tty_type[];
+
+
 char cSection[]  = "[UITPS-Color]";
 char bwSection[] = "[UITPS-Monochrome]";
 int processes;
@@ -108,14 +115,15 @@ configuration *config;
 char *ps_vect[MAX_PROCESS];
 int first_on_screen, current_process;
 window *title_win, *header_win, *screen_win, *status_win;
-static char title_text[]  = " UNIX Interactive Tools 4.1c - Process Viewer/Killer";
-static char status_text[] = " Arrows-MOVE  space-SIGNALS  r-REFRESH  ^C/F10-QUIT   ENTER-KILL";
+static char title_text[]  =
+    " UNIX Interactive Tools 4.2c - Process Viewer/Killer";
+static char UitPsModeHelp[256];
 static char no_perm[]     = "not owner !";
 static char no_proc[]     = "no such process ! (REFRESH recommended)";
 
 struct SIGNAL
 {
-    char signame[16];
+    char signame[8];
     int  signal;
 };
 
@@ -142,10 +150,72 @@ static struct SIGNAL signals[] =
 };
 
 
+#define BUILTIN_OPERATIONS	25
+
+
+#define BUILTIN_CursorUp	0
+#define BUILTIN_CursorDown	1
+#define BUILTIN_PageUp		2
+#define BUILTIN_PageDown        3
+#define BUILTIN_Home		4
+#define BUILTIN_End		5
+#define BUILTIN_NextSignal	6
+#define BUILTIN_SIGHUP		7
+#define BUILTIN_SIGINT		8
+#define BUILTIN_SIGQUIT		9
+#define BUILTIN_SIGILL		10
+#define BUILTIN_SIGFPE		11
+#define BUILTIN_SIGKILL		12
+#define BUILTIN_SIGUSR1		13
+#define BUILTIN_SIGSEGV		14
+#define BUILTIN_SIGUSR2		15
+#define BUILTIN_SIGPIPE		16
+#define BUILTIN_SIGALRM		17
+#define BUILTIN_SIGTERM		18
+#define BUILTIN_SIGCHLD		19
+#define BUILTIN_SIGCONT		20
+#define BUILTIN_KillProcess	21
+#define BUILTIN_Refresh		22
+#define BUILTIN_Exit		23
+#define BUILTIN_HardRefresh	24
+
+
+#define MAX_BUILTIN_NAME	15
+
+char built_in[BUILTIN_OPERATIONS][MAX_BUILTIN_NAME] =
+{
+    "<CursorUp>",
+    "<CursorDown>",
+    "<PageUp>",
+    "<PageDown>",
+    "<Home>",
+    "<End>",
+    "<NextSignal>",
+    "<SIGHUP>",
+    "<SIGINT>",
+    "<SIGQUIT>",
+    "<SIGILL>",
+    "<SIGFPE>",
+    "<SIGKILL>",
+    "<SIGUSR1>",
+    "<SIGSEGV>",
+    "<SIGUSR2>",
+    "<SIGPIPE>",
+    "<SIGALRM>",
+    "<SIGTERM>",
+    "<SIGCHLD>",
+    "<SIGCONT>",
+    "<KillProcess>",
+    "<Refresh>",
+    "<Exit>",
+    "<HardRefresh>",
+};
+
+
 void removelog(void)
 {
     char name[PATH_MAX];
-    
+
     sprintf(name, "%s/uitps.1.%d", TempDirectory, getpid());
     remove(name);
     sprintf(name, "%s/uitps.2.%d", TempDirectory, getpid());
@@ -164,7 +234,7 @@ void settitle(void)
     tty_foreground(TitleForeground);
     tty_background(TitleBackground);
 
-    window_cursormove(title_win, 0, 0);
+    window_cursormove_notify(title_win, 0, 0);
     window_write(temp, SCREEN_X);
 }
 
@@ -180,7 +250,7 @@ void setheader(void)
     tty_foreground(HeaderForeground);
     tty_background(HeaderBackground);
 
-    window_cursormove(header_win, 0, 0);
+    window_cursormove_notify(header_win, 0, 0);
     window_write(temp, SCREEN_X);
 }
 
@@ -188,32 +258,33 @@ void setheader(void)
 void setstatus(char *what)
 {
     char temp[256];
-    
+
     memset(temp, ' ', SCREEN_X);
     if (what)
 	memcpy(temp, what, strlen(what));
     else
-	memcpy(temp, status_text, strlen(status_text));
+	memcpy(temp, UitPsModeHelp, strlen(UitPsModeHelp));
 
     tty_bright(StatusBrightness);
     tty_foreground(StatusForeground);
     tty_background(StatusBackground);
 
-    window_cursormove(status_win, 0, 0);
-    window_write(temp, SCREEN_X);	
+    window_cursormove_notify(status_win, 0, 0);
+    window_write(temp, SCREEN_X - (sizeof(signals[0].signame) - 1) - 1);
 }
 
 
 void setsignal(void)
 {
-    int len = strlen(signals[signal_type].signame);
+    int len = sizeof(signals[0].signame) - 1;
 
     tty_bright(StatusBrightness);
     tty_foreground(StatusForeground);
     tty_background(StatusBackground);
 
-    window_cursormove(status_win, 0, SCREEN_X - len - 1);
+    window_cursormove_notify(status_win, 0, SCREEN_X - len - 1);
     window_write(signals[signal_type].signame, len);
+    window_putch(' ');
 }
 
 
@@ -320,7 +391,7 @@ void update_process(int process, int update_color)
 	tty_bright(ScreenBrightness);
     }
 
-    window_cursormove(screen_win, process - first_on_screen, 0);
+    window_cursormove_notify(screen_win, process - first_on_screen, 0);
     window_write(temp, SCREEN_X);
 }
 
@@ -328,20 +399,33 @@ void update_process(int process, int update_color)
 void update_all(void)
 {
     int i;
-    
+    char fill[256];
+
     tty_foreground(ScreenForeground);
     tty_background(ScreenBackground);
     tty_bright(ScreenBrightness);
     window_cursormove(screen_win, 0, 0);
+    memset(fill, ' ', SCREEN_X);
 
     for (i = first_on_screen;
          i < processes && (i - first_on_screen < SCREEN_Y - 3); i++)
 	    if (i != current_process)
 		update_process(i, OFF);
 	    else
-		window_cursormove(screen_win, i - first_on_screen + 1, 0);
+		window_cursormove_notify(screen_win, i - first_on_screen, 0);
 
     update_process(current_process, ON);
+
+    tty_background(ScreenBackground);
+    tty_bright(ScreenBrightness);
+
+    for (; i - first_on_screen < SCREEN_Y - 3; i++)
+    {
+	window_cursormove_notify(screen_win, i - first_on_screen, 0);
+	window_write(fill, SCREEN_X);
+    }
+
+    window_cursormove(screen_win, current_process-first_on_screen, SCREEN_X-1);
 }
 
 
@@ -357,6 +441,7 @@ void clean_up(void)
 
     tty_end();
     tty_clrscr();
+    tty_exit();
     removelog();
 }
 
@@ -365,7 +450,6 @@ void fatal(char *postmsg)
 {
     clean_up();
     fprintf(stderr, "Fatal error : %s !\n", postmsg);
-    removelog();
     exit(1);    
 }
 
@@ -412,31 +496,36 @@ void panic(int signum)
     switch (signum)
     {
 	case SIGHUP:
+
 	    clean_up();
 	    /* :-) */
 	    fprintf(stderr, "Got SIGHUP !\n");
 	    break;
 
 	case SIGINT:
+
 	    clean_up();
 	    break;
 
 	case SIGTERM:
 	case SIGQUIT:
+
 	    clean_up();
 	    fprintf(stderr, "Got %s !\nbye\n",
 	    	   (signum == SIGTERM) ? "SIGTERM" : "SIGQUIT");
 	    break;
 
 	case SIGSEGV:
+
 	    tty_end();
 	    tty_clrscr();
+	    tty_exit();
+	    removelog();
 	    fprintf(stderr, "Got SIGSEGV !\n");
 	    fprintf(stderr, "Please report to tudor@ulise.cs.pub.ro\n");
 	    break;
     }
 
-    removelog();
     exit(1);
 }
 
@@ -444,13 +533,17 @@ void panic(int signum)
 int main(int argc, char *argv[])
 {
     FILE *stdout_log;
+    struct key_struct *ks;
     int key, repeat_count;
     char tmp[256], buf[256];
-    int sectionptr = 0, i, index;
+    int sectionptr = 0, i, j, index, first_time = 1;
     char *home, *data = NULL, fname[PATH_MAX], *section;
     int need_update, need_update_all, old_current_process;
 
-    
+
+    termcap_init();
+    tty_kbdinit(1);
+
     signal(SIGTERM, panic);
     signal(SIGINT , panic);
     signal(SIGQUIT, panic);
@@ -464,8 +557,6 @@ int main(int argc, char *argv[])
     signal(SIGUSR2, SIG_IGN);
     signal(SIGTSTP, SIG_IGN);
     signal(SIGCONT, SIG_IGN);
-    signal(SIGTTIN, SIG_IGN);
-    signal(SIGTTOU, SIG_IGN);
     signal(SIGALRM, SIG_IGN);
     signal(SIGPIPE, SIG_IGN);
     signal(SIGFPE,  SIG_IGN);
@@ -476,55 +567,116 @@ int main(int argc, char *argv[])
         strcpy(configfile, home);
         if (configfile[strlen(configfile) - 1] != '/') strcat(configfile, "/");
     }
-    strcat(configfile, ".uitrc");
+    strcat(configfile, ".uitrc.");
+    strcat(configfile, tty_type);
 
     config = configuration_init(configfile);
- 
+
     if (configuration_getstatus(config) == STATUS_OK)
     {
-        if ((sectionptr = configuration_getsectionptr(config, "[Setup]")) == -1)
+	sectionptr = configuration_getsectionptr(config,"[Setup]");
+        if (sectionptr == -1)
 	    puts("[Setup] section missing.");
         else
         {
-	    configuration_getfielddata(config, sectionptr, "TempDirectory", &data, 1, DO_SEEK);
+	    configuration_getfielddata(config, sectionptr, "TempDirectory",
+				       &data, 1, DO_SEEK);
 	    strncpy(TempDirectory, data ? data : homedirectory, PATH_MAX);
 #ifdef HAVE_LINUX
-	    configuration_getfielddata(config, sectionptr, "LinuxConsole", &data, 1, DO_SEEK);
+	    configuration_getfielddata(config, sectionptr, "LinuxConsole",
+				       &data, 1, DO_SEEK);
 	    LinuxConsole = data && strcmp(data, "ON") == 0;
-	    configuration_getfielddata(config, sectionptr, "ColorMonitor", &data, 1, DO_SEEK);
+	    configuration_getfielddata(config, sectionptr, "ColorMonitor",
+				       &data, 1, DO_SEEK);
 	    ColorMonitor = data && strcmp(data, "ON") == 0;
 #endif
 	}
 
+	sectionptr = configuration_getsectionptr(config, "[UITPS-Setup]");
+        if (sectionptr == -1)
+	    fprintf(stderr, "[UITPS-Setup] section missing.");
+	else
+	{
+	    configuration_getfielddata(config, sectionptr,
+				       "UitPsModeHelp",
+				       &data, 1, DO_SEEK);
+	    if (data) strncpy(UitPsModeHelp, data, 256);
+	}
+
 	section = (LinuxConsole && ColorMonitor) ? cSection : bwSection;
 
-	if ((sectionptr = configuration_getsectionptr(config, section)) == -1)
+	sectionptr = configuration_getsectionptr(config, section);
+	if (sectionptr == -1)
 	    fprintf(stderr, "%s section missing.", section);
 	else
 	    for (i = 0; i < KILLER_FIELDS; i++)
 	    {
-		configuration_getfielddata(config, sectionptr, PSFields[i], &data, 1, DO_SEEK);
+		configuration_getfielddata(config, sectionptr, PSFields[i],
+					   &data, 1, DO_SEEK);
         	if (!data || (index  = tty_getcolorindex(data))== -1)
                     fprintf(stderr, "Invalid %s (%s).\n", PSFields[i], data);
 		else
 		    PSColors[i] = index;
 	    }
+
+        sectionptr = configuration_getsectionptr(config, "[UITPS-Keys]");
+        if (sectionptr == -1)
+	    fprintf(stderr, "[UITPS-Keys] section missing.");
+	else
+	{
+	    char key_seq[16];
+	    char *contents;
+
+	    for (i = 0; i < MAX_KEYS; i++)
+	    {
+	        configuration_getfielddata(config, sectionptr, key_seq,
+					   &contents, 1, NO_SEEK);
+
+		if (*key_seq == 0) break;
+
+		if (contents == NULL) continue;
+
+	        for (j = 0; j < BUILTIN_OPERATIONS; j++)
+	            if (strcmp(contents, built_in[j]) == 0)
+	                break;
+
+		if (j < BUILTIN_OPERATIONS)
+		{
+		    if (tty_key_convert(key_seq))
+		        tty_key_list_insert(key_seq, (void *)j);
+		}
+		else
+		    fprintf(stderr, "Invalid built-in operation: %s\n",
+			    contents);
+	    }
+
+	    if (i == MAX_KEYS)
+	        fprintf(stderr,
+			"Too many key sequences. Only %d are allowed.\n",
+			MAX_KEYS);
+	}
 	configuration_end(config);
     }
     else
-        fprintf(stderr, "Cannot open configuration file %s.\nUsing defaults.\n", configfile);
+    {
+        fprintf(stderr, "\nCannot open configuration file %s\n", configfile);
+	fprintf(stderr, "See the manual page for details.\n");
+	fprintf(stderr, "If your TERM environment variable is, let's say,\n");
+	fprintf(stderr, "vt102, your configuration file name should be\n");
+	fprintf(stderr, ".uitrc.vt102 . Try renaming .uitrc.console ...\n");
+	exit(1);
+    }
 
     tty_getsize(&SCREEN_X, &SCREEN_Y);
+    tty_startup();
 
-    header_text = malloc(SCREEN_X + 1);
+    header_text = (char *)malloc(SCREEN_X + 1);
 
     title_win  = window_init(0, 0, 	      1,	    SCREEN_X);
     header_win = window_init(0, 1, 	      1,	    SCREEN_X);
-    screen_win = window_init(0, 2,            SCREEN_Y - 2, SCREEN_X);
+    screen_win = window_init(0, 2,            SCREEN_Y - 3, SCREEN_X);
     status_win = window_init(0, SCREEN_Y - 1, 1,	    SCREEN_X);
 
-    termcap_init();
-    tty_kbdinit();
     tty_init();
 
     first_on_screen = current_process = 0;
@@ -551,10 +703,14 @@ restart:
     build_ps_list(stdout_log);
     fclose(stdout_log);
 
-    tty_foreground(ScreenForeground);
-    tty_background(ScreenBackground);
-    tty_bright(ScreenBrightness);
-    tty_clrscr();
+    if (first_time)
+    {
+	tty_foreground(ScreenForeground);
+	tty_background(ScreenBackground);
+	tty_bright(ScreenBrightness);
+	tty_clrscr();
+	first_time = 0;
+    }
 
     settitle();
     setstatus(NULL);
@@ -565,13 +721,15 @@ restart:
 
     update_all();
 
-    while ((key = tty_getkey(&repeat_count)), 1)
+    while (1)
     {
-        if (isupper(key)) key = tolower(key);
+	ks  = tty_getkey(&repeat_count);
+	key = (int)ks->aux_data;
+
 	switch (key)
 	{
-	    case -(188 + TTY_FIRST_CONFIG_KEY):		/* hard coded ^y */
-	    case -TTY_CURSOR_UP:
+	    case BUILTIN_CursorUp:
+
 	        need_update_all = need_update = 0;
 	        while (repeat_count--)
 	        {
@@ -596,8 +754,8 @@ restart:
 	                update_process(current_process, ON);
 	        break;
 
-	    case -(22 + TTY_FIRST_CONFIG_KEY):		/* hard coded ^b */
-	    case -TTY_CURSOR_DOWN:
+	    case BUILTIN_CursorDown:
+
 	        need_update_all = need_update = 0;
 	        while (repeat_count--)
 	        {
@@ -620,9 +778,8 @@ restart:
 	                update_process(current_process, ON);
 	        break;
 
-	    case '-':
-	    case -TTY_CURSOR_LEFT:
-	    case -TTY_PGUP:
+	    case BUILTIN_PageUp:
+
 	        if (current_process == 0) break;
 
 		old_current_process = current_process;
@@ -645,9 +802,8 @@ restart:
 
 	        break;
 
-	    case '+':
-	    case -TTY_CURSOR_RIGHT:
-	    case -TTY_PGDN:
+	    case BUILTIN_PageDown:
+
 	        if (current_process == processes - 1) break;
 
 	        old_current_process = current_process;
@@ -663,7 +819,9 @@ restart:
 		    else
 		    {
 			current_process += SCREEN_Y - 3;
-			first_on_screen = min(first_on_screen + SCREEN_Y - 3, (processes - 1) - (SCREEN_Y - 3) + 1);
+			first_on_screen = min(first_on_screen + SCREEN_Y - 3,
+					      (processes - 1) -
+					      (SCREEN_Y - 3) + 1);
 		    }
 
 		if (processes > SCREEN_Y - 3)
@@ -676,49 +834,57 @@ restart:
 
 	        break;
 
-	    case -(72 + TTY_FIRST_CONFIG_KEY):		/* hard coded ^ku */
-	    case -TTY_HOME:
+	    case BUILTIN_Home:
+
 	        if (current_process == 0) break;
 	        current_process = first_on_screen = 0;
 	        update_all();
 	    	break;
 
-	    case -(73 + TTY_FIRST_CONFIG_KEY):		/* hard coded ^kv */
-	    case -TTY_END:
+	    case BUILTIN_End:
+
 	        if (current_process == processes - 1) break;
 	        current_process = processes - 1;
 	        first_on_screen = max(0, (processes - 1) - (SCREEN_Y - 3) + 1);
 	        update_all();
 	    	break;
 
-	    case ' ':
+	    case BUILTIN_NextSignal:
+
 	        signal_type++;
 	        signal_type %= sizeof(signals) / sizeof(struct SIGNAL);
 	        setsignal();
 	        break;
 
-	    case 'h': signal_type =  0; setsignal(); break;	/* SIGHUP  */
-	    case 'i': signal_type =  1; setsignal(); break;	/* SIGINT  */
-	    case 'q': signal_type =  2; setsignal(); break;	/* SIGQUIT */
-	    case 'j': signal_type =  3; setsignal(); break;	/* SIGILL  */
-	    case 'f': signal_type =  4; setsignal(); break;	/* SIGFPE  */
-	    case 'k': signal_type =  5; setsignal(); break;	/* SIGKILL */
-	    case '1': signal_type =  6; setsignal(); break;	/* SIGUSR1 */
-	    case 's': signal_type =  7; setsignal(); break;	/* SIGSEGV */
-	    case '2': signal_type =  8; setsignal(); break;	/* SIGUSR2 */
-	    case 'p': signal_type =  9; setsignal(); break;	/* SIGPIPE */
-	    case 'a': signal_type = 10; setsignal(); break;	/* SIGALRM */
-	    case 't': signal_type = 11; setsignal(); break;	/* SIGTERM */
-	    case 'y': signal_type = 12; setsignal(); break;	/* SIGCHLD */
-	    case 'c': signal_type = 13; setsignal(); break;	/* SIGCONT */
+	    case BUILTIN_SIGHUP : signal_type =  0; setsignal(); break;
+	    case BUILTIN_SIGINT : signal_type =  1; setsignal(); break;
+	    case BUILTIN_SIGQUIT: signal_type =  2; setsignal(); break;
+	    case BUILTIN_SIGILL : signal_type =  3; setsignal(); break;
+	    case BUILTIN_SIGFPE : signal_type =  4; setsignal(); break;
+	    case BUILTIN_SIGKILL: signal_type =  5; setsignal(); break;
+	    case BUILTIN_SIGUSR1: signal_type =  6; setsignal(); break;
+	    case BUILTIN_SIGSEGV: signal_type =  7; setsignal(); break;
+	    case BUILTIN_SIGUSR2: signal_type =  8; setsignal(); break;
+	    case BUILTIN_SIGPIPE: signal_type =  9; setsignal(); break;
+	    case BUILTIN_SIGALRM: signal_type = 10; setsignal(); break;
+	    case BUILTIN_SIGTERM: signal_type = 11; setsignal(); break;
+	    case BUILTIN_SIGCHLD: signal_type = 12; setsignal(); break;
+	    case BUILTIN_SIGCONT: signal_type = 13; setsignal(); break;
+
+	    case BUILTIN_HardRefresh:
+
+	    	first_time = 1;
+
+	    case BUILTIN_Refresh:
 
-	    case 'r': 
 	    	goto restart;
 
-	    case -TTY_F10:
+	    case BUILTIN_Exit:
+
 	        goto end;
 
-	    case KEY_ENTER:
+	    case BUILTIN_KillProcess:
+
 	        if (!kill_process(current_process))
 	        {
 	            tty_beep();
@@ -736,7 +902,7 @@ restart:
     }
 
   end:
-    
+
     clean_up();
     return 0;
 }
